"""
Imports logic for "django_expanded_test_cases/mixins/" folder.
Makes project imports to this folder behave like a standard file.
"""

# Expanded "Base" TestCase utility mixins.
from .core_mixin import CoreTestCaseMixin


# Expanded "Integration" TestCase utility mixins.
from .response_mixin import ResponseTestCaseMixin
