
<head>
  <meta charset="utf-8">
  <title>{{ header }} | Test Views</title>
</head>


<body>
  {% if messages %}
    <ul>
      {% for message in messages %}
        <li><p>
          {{ message }}
        </p></li>
      {% endfor %}
    </ul>
  {% endif %}

  {% if header %}
    <h1>{{ header }} Header</h1>
  {% endif %}

  {% if text %}
    <p>{{ text }}</p>
  {% endif %}

  {% if li_set %}
    <ul>
      {% for li_item in li_set %}
        <li><p>{{ li_item}}</p></li>
      {% endfor %}
    </ul>
  {% endif %}
</body>
